"""Access ubank's API using Python.

Run as a script to create a new passkey.

The Api class is a ubank API client.

The Client class is lower level HTTP client that uses passkey authentication.
"""

from __future__ import annotations

import argparse
import json
import logging
import time
import uuid
from base64 import b64encode, urlsafe_b64encode
from datetime import date, datetime
from decimal import Decimal
from getpass import getpass
from typing import IO, AnyStr, Optional

import httpx
import soft_webauthn
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from fido2 import cbor
from meatie import endpoint
from meatie_httpx import Client as MeatieClient
from pydantic import BaseModel, Field
from soft_webauthn import SoftWebauthnDevice

__version__ = "2.1.0"


# Unchanging headers in every request.
base_headers = {
    "Origin": "ionic://bank86400",
    "x-api-version": "32",
    "x-private-api-key": "ANZf5WgzmVLmTUwAQyuCq7LspXF2pd4N",
}

# Referenced in Client and add_passkey() for attestation and assertion.
origin = "https://www.ubank.com.au"


class Address(BaseModel):
    addressFormat: Optional[str]
    addressType: Optional[str]
    flatOrBoxNumber: Optional[str]
    flatOrBoxType: Optional[str]
    postcode: Optional[str]
    propertyName: Optional[str]
    state: Optional[str]
    streetName: Optional[str]
    streetNumber: Optional[str]
    streetType: Optional[str]
    suburb: Optional[str]


class CustomerDetails(BaseModel):
    addresses: list[Address]
    countriesOfCitizenship: list[str]
    customerId: Optional[str]
    dateCreated: date
    dateOfBirth: date
    email: Optional[str]
    emailVerified: bool
    fatcaCrsProvided: bool
    firstName: Optional[str]
    middleName: Optional[str]
    lastName: Optional[str]
    additionalNames: Optional[str]
    mobileNumber: Optional[str]
    externalCustomerNumber: Optional[str]
    jurisdictionSourceOfWealth: list
    natureAndPurposeOfRelationship: list
    nonResidentTaxDetails: list
    occupationId: Optional[str]
    title: Optional[str]
    userId: str
    financialCrime: dict
    editAddressEnabled: bool


class Balance(BaseModel):
    currency: str
    current: float
    available: float


class Account(BaseModel):
    id: str
    number: str
    bsb: str
    label: str
    nickname: str
    type: str
    balance: Balance
    status: str
    lastBalanceRefresh: datetime
    openDate: datetime
    isJointAccount: bool
    depositProductData: dict
    metadata: Optional[dict] = None


class LinkedBank(BaseModel):
    bankId: int
    shortBankName: str
    accounts: list[Account]


class AccountsResponse(BaseModel):
    linkedBanks: list[LinkedBank]


class Value(BaseModel):
    amount: Decimal
    currency: str


class From(BaseModel):
    name: Optional[str] = None
    legalName: Optional[str] = None
    bsb: Optional[str] = None
    number: Optional[str] = None
    description: Optional[str] = None


To = From


class Transaction(BaseModel):
    id: str
    cbsId: str
    bankId: str
    accountId: str
    posted: datetime
    completed: datetime
    value: Value
    type: str
    shortDescription: str
    # Everything else seems optional.
    narration: Optional[Value] = None
    balance: Optional[Value] = None
    debitOrCredit: Optional[str] = None
    description: Optional[str] = None
    category: Optional[str] = None
    paymentScheme: Optional[str] = None
    receiptNumber: Optional[str] = None
    from_: Optional[From] = Field(default=None, alias="from")
    bpayBiller: Optional[dict] = None
    to: Optional[To] = None
    walletType: Optional[str] = None
    cardNumber: Optional[str] = None
    nppTransactionId: Optional[str] = None
    nppServiceOverlay: Optional[str] = None
    nppCategoryPurposeCode: Optional[str] = None
    nppCreditorReference: Optional[str] = None
    nppIdentification: Optional[str] = None
    nppSchemeName: Optional[str] = None
    terminalId: Optional[str] = None
    systemTraceAuditNumber: Optional[str] = None
    visaTerminalId: Optional[str] = None
    typeCode: Optional[str] = None
    longDescription: Optional[str] = None
    lwc: Optional[dict] = None


class TransactionsSearchResponse(BaseModel):
    nextPageId: str
    totalCount: int
    totalAmount: str
    transactions: list[Transaction]


class TransactionsSearchBody(BaseModel):
    timezone: str = "Australia/Lord_Howe"
    fromDate: date
    toDate: date
    limit: int = 5


class TransactionsResponse(BaseModel):
    nextPageId: str
    transactions: list[Transaction]
    pendingTransactions: list


class Device(BaseModel):
    type: str
    deviceUuid: str
    deviceName: str
    deviceCreatedOn: Optional[str] = None
    dateCreated: int
    dateCreatedTimestamp: datetime
    enabled: bool
    isEditable: bool


class Card(BaseModel):
    accountIds: list[str]
    cardToken: str
    cardNumber: str
    bankId: int
    cardId: str
    nameOnCard: str
    cardStatus: str
    expiryDate: str
    cardType: str
    panReferenceId: str
    locked: bool
    cardArtName: str
    lastProductionDate: datetime
    cardControls: list[dict]


class CardsResponse(BaseModel):
    cards: list[Card]
    cardReplacements: list


class Passkey:
    """Represents a passkey registered with ubank."""

    def __init__(self, name: str):
        """name is passkey/device name shown in emails and app."""
        super().__init__()
        self.name = name
        # Generate a fresh hardware ID.
        hardware_id = str(uuid.uuid4())
        # Start with an empty device ID. ubank will assign one later.
        device_id = ""
        # Hard coded Build device meta string from app version and device name.
        device_meta = json.dumps(
            {
                # Keep track with the latest version in case API blocks old versions: https://apps.apple.com/au/app/id1449543099.
                "appVersion": "11.103.3",
                "binaryVersion": "11.103.3",
                # Keep track with latest iPhone model in case API blocks old models.
                # Matches format of "Hardware strings" row in this table, replacing ',' with '-':
                # https://en.wikipedia.org/wiki/List_of_iPhone_models
                "deviceName": "iPhone17-3",
                "environment": "production",
                "instance": "live",
                "native": True,
                "platform": "ios",
            }
        )
        # Start with empty username, it will be assigned by ubank later.
        username = ""
        self.name = name
        self.hardware_id = hardware_id
        self.device_id = device_id
        self.device_meta = device_meta
        self.username = username
        self.soft_webauthn_device = UserVerifiedDevice()

    def dump(self, file: IO[bytes], password: str = ""):
        """Serializes passkey to file, encrypted with a password.

        Uses the following as a guide for password encryption: https://cryptography.io/en/latest/fernet/#using-passwords-with-fernet

        This uses a hardcoded salt because I'm not interested in additionally keeping
        track of unique salts. A password alone will have to be good enough!

        AI suggests generating random hardcoded salts, derived from a hash of password.
        But I'm hard coding it to b"". Not sure random hardcoded salts give benefit
        in this case. If someone knows your password, you're popped!

        Not supplying a password encrypts the file using a key derived from an empty string.
        """
        fernet = Fernet(derive_key(password))
        file.write(
            fernet.encrypt(
                cbor.encode(
                    {
                        "name": self.name,
                        "hardware_id": self.hardware_id,
                        "device_id": self.device_id,
                        "device_meta": self.device_meta,
                        "username": self.username,
                        "soft_webauthn_device_dict": to_dict(self.soft_webauthn_device),
                    }
                )
            )
        )

    @classmethod
    def load(cls, file: IO[AnyStr], password: str = "") -> Passkey:
        """Deserializes passkey from `file`, decrypted with `password`."""
        fernet = Fernet(derive_key(password))
        deserialized_passkey = cbor.decode(fernet.decrypt(file.read()))
        passkey = Passkey(deserialized_passkey["name"])
        passkey.hardware_id = deserialized_passkey["hardware_id"]
        passkey.device_id = deserialized_passkey["device_id"]
        passkey.device_meta = deserialized_passkey["device_meta"]
        passkey.username = deserialized_passkey["username"]
        passkey.soft_webauthn_device = from_dict(
            deserialized_passkey["soft_webauthn_device_dict"]
        )
        return passkey


class UserVerifiedDevice(SoftWebauthnDevice):
    """Software webauthn device with UV bit flag ALWAYS set."""

    def create(self, options, origin):
        """IDENTICAL to super().create(), except User Verification flag set."""
        if {"alg": -7, "type": "public-key"} not in options["publicKey"][
            "pubKeyCredParams"
        ]:
            raise ValueError(
                "Requested pubKeyCredParams does not contain supported type"
            )

        if ("attestation" in options["publicKey"]) and (
            options["publicKey"]["attestation"] not in [None, "none"]
        ):
            raise ValueError("Only none attestation supported")

        # prepare new key
        self.cred_init(
            options["publicKey"]["rp"]["id"], options["publicKey"]["user"]["id"]
        )

        # generate credential response
        client_data = {
            "type": "webauthn.create",
            "challenge": soft_webauthn.urlsafe_b64encode(
                options["publicKey"]["challenge"]
            )
            .decode("ascii")
            .rstrip("="),
            "origin": origin,
        }

        rp_id_hash = soft_webauthn.sha256(self.rp_id.encode("ascii"))
        # Set Bit 2, User Verification (UV) also.
        # https://developer.mozilla.org/en-US/docs/Web/API/Web_Authentication_API/Authenticator_data
        flags = b"\x45"  # attested_data + user_present + user_verified
        sign_count = soft_webauthn.pack(">I", self.sign_count)
        credential_id_length = soft_webauthn.pack(">H", len(self.credential_id))
        cose_key = soft_webauthn.cbor.encode(
            soft_webauthn.ES256.from_cryptography_key(self.private_key.public_key())
        )
        attestation_object = {
            "authData": rp_id_hash
            + flags
            + sign_count
            + self.aaguid
            + credential_id_length
            + self.credential_id
            + cose_key,
            "fmt": "none",
            "attStmt": {},
        }

        return {
            "id": soft_webauthn.urlsafe_b64encode(self.credential_id),
            "rawId": self.credential_id,
            "response": {
                "clientDataJSON": json.dumps(client_data).encode("utf-8"),
                "attestationObject": soft_webauthn.cbor.encode(attestation_object),
            },
            "type": "public-key",
        }

    def get(self, options, origin):
        """IDENTICAL to super().create(), except User Verification flag set."""

        if self.rp_id != options["publicKey"]["rpId"]:
            raise ValueError("Requested rpID does not match current credential")

        self.sign_count += 1

        # prepare signature
        client_data = json.dumps(
            {
                "type": "webauthn.get",
                "challenge": soft_webauthn.urlsafe_b64encode(
                    options["publicKey"]["challenge"]
                )
                .decode("ascii")
                .rstrip("="),
                "origin": origin,
            }
        ).encode("utf-8")
        client_data_hash = soft_webauthn.sha256(client_data)

        rp_id_hash = soft_webauthn.sha256(self.rp_id.encode("ascii"))
        # Set Bit 2, User Verification (UV) also.
        # https://developer.mozilla.org/en-US/docs/Web/API/Web_Authentication_API/Authenticator_data
        flags = b"\x05"
        sign_count = soft_webauthn.pack(">I", self.sign_count)
        authenticator_data = rp_id_hash + flags + sign_count

        signature = self.private_key.sign(
            authenticator_data + client_data_hash,
            soft_webauthn.ec.ECDSA(soft_webauthn.hashes.SHA256()),
        )

        # generate assertion
        return {
            "id": soft_webauthn.urlsafe_b64encode(self.credential_id),
            "rawId": self.credential_id,
            "response": {
                "authenticatorData": authenticator_data,
                "clientDataJSON": client_data,
                "signature": signature,
                "userHandle": self.user_handle,
            },
            "type": "public-key",
        }


class Api(MeatieClient):
    """A ubank API client.

    Provides methods for interacting with the following resources:

    - customer details
    - accounts
    - transactions
    - cards
    - contacts
    - devices (passkeys)

    This is a `MeatieClient`. It has a bunch of methods... with nothing in them.
    Meatie generates code for calling endpoints automatically! It does this by inspecting
    type signatures. And using [descriptors](https://docs.python.org/3/howto/descriptor.html).

    I've heard of decorators, but not descriptors. Sounds powerful. Worth looking into.
    """

    def __init__(self, passkey: Passkey) -> None:
        super().__init__(Client(passkey))

    @endpoint("customer-details")
    def get_customer_details(self) -> CustomerDetails: ...

    @endpoint("accounts")
    def get_accounts(self) -> AccountsResponse: ...

    @endpoint("accounts/summary")
    def get_accounts_summary(self) -> AccountsResponse: ...

    @endpoint("accounts/{account_id}/bank/{bank_id}/transactions")
    def get_account_transactions(
        self,
        account_id: str,
        bank_id: str,
        customerId: str,
        limit: int = 50,
        pageId: str = "",
        query: str = "",
    ) -> TransactionsResponse: ...

    @endpoint("accounts/transactions/search")
    def post_accounts_transactions_search(
        self, body: TransactionsSearchBody
    ) -> TransactionsSearchResponse: ...

    @endpoint("cards")
    def get_cards(self) -> CardsResponse: ...

    @endpoint("v2/devices")
    def get_devices(self, deviceUuid: str) -> list[Device]: ...

    @endpoint("device/{device_id}")
    def delete_device(self, device_id: str) -> str: ...

    @endpoint("contacts")
    def get_contacts(self) -> dict: ...


class Client(httpx.Client):
    """httpx.Client initialised with a passkey to make authenticated requests to ubank.

    Initialising the class performs webauthn authentication using the supplied passkey.
    ubank is the Relying Party (RP). The supplied passkey signs an assertion using
    a challenge from RP - SoftWebauthnDevice.get().  The assertion is sent back
    to RP for verification.

    The result from this call is transformed and sent back to ubank.

    completes webauthn flow

        super().__init__(
            headers=client.headers,
            cookies=client.cookies,
            base_url="https://api.ubank.com.au/app/v1/",
        )

    Initialised with a passkey.  Client(pass)

    Requests are authenticated using the supplied passkey.

    Use this class as a context manager to ensure ubank sessions and HTTP connections
    are properly closed.

    ```python
    with Client(passkey) as client:
        ...
    ```

    `base_url` is set to https://api.ubank.com.au/app/v1/. Use relative paths in requests:

    ```python
    client.get("accounts/summary")
    ```
    """

    def __init__(self, passkey: Passkey) -> None:
        """Initialises ubank session using passkey to authenticate.

        Caught HTTPStatusErrors are re-raised with a note containing the API's error
        response text. This requires Python >= 3.11.
        """
        with httpx.Client(
            # Headers that are present from the get go. Not all
            headers={
                **base_headers,
                "x-hardware-id": passkey.hardware_id,
                "x-device-id": passkey.device_id,
                "x-device-meta": passkey.device_meta,
            },
        ) as client:
            # Hack signature counter to Unix time. This 32-bit counter value can
            # be incremented by *any* positive value. By using Unix time, we don't
            # have to muck about keeping track of counter values in the passkey file.
            # https://www.w3.org/TR/webauthn-2/#signature-counter
            passkey.soft_webauthn_device.sign_count = int(time.time())

            # Initiate authentication flow to receive challenge from relying party
            # (ubank).
            try:
                response = client.get(
                    "https://api.ubank.com.au/app/v1/session/authorize",
                    params={
                        "username": passkey.username,
                    },
                ).raise_for_status()
            except httpx.HTTPStatusError as e:
                e.add_note(e.response.text)
                raise

            # Parse credential request options from response.
            options = parse_public_key_credential_request_options(
                response.json()["publicKeyCredentialRequestOptions"]
            )
            # Make assertion object suitable for ubank by making values JSON-serializable.
            assertion = prepare_assertion(
                passkey.soft_webauthn_device.get(options, origin)
            )
            # Complete authentication flow by sending signed assertion to relying
            # party.
            try:
                response = client.post(
                    "https://api.ubank.com.au/app/v1/challenge/fido2-assertion",
                    # Query parameters come from previous response.
                    params={
                        "nonce": response.json()["nonce"],
                        "state": response.json()["state"],
                        "session": response.json()["session"],
                    },
                    json={
                        "assertion": json.dumps(assertion),
                        # flowID comes from previous response.
                        "flowId": response.json()["flowId"],
                        "origin": origin,
                    },
                    # Set access and auth token headers from previous response.
                    headers={
                        "x-access-token": response.json()["accessToken"],
                        "x-auth-token": response.json()["accessToken"],
                    },
                ).raise_for_status()
            except httpx.HTTPStatusError as e:
                e.add_note(e.response.text)
                raise
            # Set access and auth token headers for future requests.
            self.access_token = client.headers["x-access-token"] = client.headers[
                "x-auth-token"
            ] = response.json()["accessToken"]
            # Store other tokens in order to kill session in future.
            self.refresh_token = response.json()["refreshToken"]
            self.session_token = response.json()["sessionToken"]

        # Initialise new httpx.Client, keeping headers and cookies from client.
        super().__init__(
            headers=client.headers,
            cookies=client.cookies,
            base_url="https://api.ubank.com.au/app/v1/",
        )

    def _delete_session(self) -> None:
        """Kills ubank session."""
        self.request(
            "DELETE",
            "sessions",
            json={
                "accessToken": self.access_token,
                "refreshToken": self.refresh_token,
                "sessionToken": self.session_token,
            },
        ).raise_for_status()

    def close(self) -> None:
        """Kills ubank session before closing."""
        self._delete_session()
        super().close()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Handle use in context manager Kills ubank session before exiting the context."""
        self._delete_session()
        super().__exit__(exc_type, exc_value, traceback)


def derive_key(password: str, salt=b"") -> bytes:
    """Returns key derived from from password.

    https://cryptography.io/en/latest/fernet/#using-passwords-with-fernet
    """
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=1_000_000,
    )
    return urlsafe_b64encode(kdf.derive(password.encode("utf-8")))


def int8array_to_bytes(array: list[int]) -> bytes:
    """Converts Javascript Int8Array to unsigned bytes."""
    return b"".join(
        # length and byteorder must be specified for Python < 3.11.
        int8.to_bytes(length=1, byteorder="big", signed=True)
        for int8 in array
    )


def parse_public_key_credential_creation_options(string: str) -> dict:
    """Returns SoftWebauthnDevice create options dict parsed from ubank's
    publicKeyCredentialCreationOptions."""
    # Deserialize string.
    options = {"publicKey": json.loads(string)}
    # Convert Int8Arrays to bytes.
    options["publicKey"]["user"]["id"] = int8array_to_bytes(
        options["publicKey"]["user"]["id"]
    )
    options["publicKey"]["challenge"] = int8array_to_bytes(
        options["publicKey"]["challenge"]
    )
    for credential in options["publicKey"]["excludeCredentials"]:
        credential["id"] = int8array_to_bytes(credential["id"])
    # Fix alg values; should be int not string.
    # https://developer.mozilla.org/en-US/docs/Web/API/PublicKeyCredentialCreationOptions#alg
    for param in options["publicKey"]["pubKeyCredParams"]:
        param["alg"] = int(param["alg"])

    return options


def parse_public_key_credential_request_options(string: str) -> dict:
    """Returns SoftWebauthnDevice get options dict parsed from ubank's
    publicKeyCredentialRequestOptions."""
    # Deserialize string.
    options = {"publicKey": json.loads(string)}
    # Convert Int8Arrays to bytes.
    options["publicKey"]["challenge"] = int8array_to_bytes(
        options["publicKey"]["challenge"]
    )
    for credential in options["publicKey"]["allowCredentials"]:
        credential["id"] = int8array_to_bytes(credential["id"])

    return options


def prepare_attestation(attestation: dict) -> dict:
    """Creates JSON-serializable attestation from SoftWebauthnDevice attestation object."""
    return {
        # id is base64 bytes, decode to base64 string.
        "id": attestation["id"].decode("ascii"),
        # rawId is bytes, convert to base64 encoded string.
        "rawId": b64encode(attestation["rawId"]).decode("ascii"),
        # clientDataJSON and attestationObject are bytes, convert to base64 encoded
        # strings.
        "response": {
            key: b64encode(value).decode("ascii")
            for key, value in attestation["response"].items()
        },
        "type": attestation["type"],
    }


def prepare_assertion(assertion: dict) -> dict:
    """Creates JSON-serializable assertion from SoftWebauthnDevice assertion object."""
    return {
        # id is base64 bytes, decode to base64 string.
        "id": assertion["id"].decode("ascii"),
        # rawId is bytes, convert to base64 encoded string.
        "rawId": b64encode(assertion["rawId"]).decode("ascii"),
        # clientDataJSON, attestationObject, signature and userHandle are bytes,
        # convert to base64 encoded strings.
        "response": {
            key: b64encode(value).decode("ascii")
            for key, value in assertion["response"].items()
        },
        "type": assertion["type"],
    }


def add_passkey(username: str, password: str, passkey_name: str) -> Passkey:
    """Returns new passkey registered with ubank after prompting for security code
    sent to mobile.

    This function returns sensitive key material. You are responsible for securing
    it!

    :param username: ubank username
    :param password: ubank password
    :param passkey_name: Set passkey name (shown in ubank app)
    :return: New passkey
    """
    # Initialise a software-based passkey.
    passkey = Passkey(name=passkey_name)

    # Use a Client to persist cookies and setting default headers.
    with httpx.Client(
        headers={
            **base_headers,
            "x-hardware-id": passkey.hardware_id,
            "x-device-id": passkey.device_id,
            "x-device-meta": passkey.device_meta,
        }
    ) as client:
        # Start enrolment by identifying ourselves.
        try:
            response = client.post(
                url="https://api.ubank.com.au/app/v1/welcome",
                json={"identity": username},
            ).raise_for_status()
        except httpx.HTTPStatusError as e:
            e.add_note(e.response.text)
            raise

        # Next, authenticate with password.
        try:
            response = client.post(
                url="https://api.ubank.com.au/app/v1/challenge/password",
                json={"deviceName": passkey_name, "password": password},
                # Set access and auth token headers from previous response.
                headers={
                    "x-access-token": response.json()["accessToken"],
                    "x-auth-token": response.json()["accessToken"],
                },
            ).raise_for_status()
        except httpx.HTTPStatusError as e:
            e.add_note(e.response.text)
            raise

        # Authenticate with second factor: a security code sent to mobile.
        try:
            otp_response = client.post(
                url="https://api.ubank.com.au/app/v1/challenge/otp",
                # Set parameters returned in previous response.
                params={
                    "nonce": response.json()["nonce"],
                    "state": response.json()["state"],
                    "session": response.json()["session"],
                },
                json={
                    # flowID comes from previous response.
                    "flowId": response.json()["flowId"],
                    # Prompt interactively for security code.
                    "otpValue": input(
                        f"Enter security code sent to {response.json()['maskedMobileNumber']}: "
                    ),
                },
                # Set access and auth token headers from previous response.
                headers={
                    "x-access-token": response.json()["accessToken"],
                    "x-auth-token": response.json()["accessToken"],
                },
            ).raise_for_status()
        except httpx.HTTPStatusError as e:
            e.add_note(e.response.text)
            raise
        # Store username UUID assigned by ubank contained in this response.
        passkey.username = otp_response.json()["username"]
        # Set access and auth token headers for future requests.
        client.headers["x-access-token"] = client.headers["x-auth-token"] = (
            otp_response.json()["accessToken"]
        )

        # Initiate registration of new credential (passkey) with relying party (ubank).
        try:
            response = client.post(
                url="https://api.ubank.com.au/app/v1/v2/device",
                json={"deviceName": passkey_name, "type": "FIDO2"},
            ).raise_for_status()
        except httpx.HTTPStatusError as e:
            e.add_note(e.response.text)
            raise
        # This response contains a device ID assigned by ubank. It's not set in
        # headers just quite yet though.
        passkey.device_id = response.json()["deviceId"]
        # Parse credential creation options from response.
        options = parse_public_key_credential_creation_options(
            response.json()["publicKeyCredentialCreationOptions"]
        )
        # Make attestation object suitable for ubank by making values JSON-serializable.
        attestation = prepare_attestation(
            passkey.soft_webauthn_device.create(options, origin)
        )

        # Send public key credential attestation to relying party (ubank).
        try:
            client.post(
                url=f"https://api.ubank.com.au/app/v1/v2/device/{passkey.device_id}/activate",
                json={
                    "attestation": json.dumps(attestation),
                    "origin": origin,
                    "type": "FIDO2",
                },
            ).raise_for_status()
        except httpx.HTTPStatusError as e:
            e.add_note(e.response.text)
            raise

        # Set device ID header for future requests.
        client.headers["x-device-id"] = passkey.device_id

        # Clear this session's tokens.
        client.request(
            # Can't use client.delete() because bodies in DELETE requests have no
            # defined semantics.
            method="DELETE",
            url="https://api.ubank.com.au/app/v1/sessions",
            # Tokens were returned in /challenge/otp response.
            json={
                "accessToken": otp_response.json()["accessToken"],
                "refreshToken": otp_response.json()["refreshToken"],
                "sessionToken": otp_response.json()["sessionToken"],
            },
        )

        return passkey


def to_dict(device: UserVerifiedDevice) -> dict:
    """Converts SoftWebauthnDevice instance to dict with serialized private key."""
    serialized_private_key = (
        device.private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        if device.private_key
        else b""
    )
    return {
        "credential_id": device.credential_id,
        "serialized_private_key": serialized_private_key,
        "aaguid": device.aaguid,
        "rp_id": device.rp_id,
        "user_handle": device.user_handle,
        "sign_count": device.sign_count,
    }


def from_dict(device_dict: dict) -> UserVerifiedDevice:
    """Returns device instantiated from dict."""
    device = UserVerifiedDevice()
    device.credential_id = device_dict["credential_id"]
    device.private_key = serialization.load_pem_private_key(
        device_dict["serialized_private_key"],
        password=None,
        backend=default_backend(),
    )
    device.aaguid = device_dict["aaguid"]
    device.rp_id = device_dict["rp_id"]
    device.user_handle = device_dict["user_handle"]
    device.sign_count = device_dict["sign_count"]
    return device


def cli():
    parser = argparse.ArgumentParser(
        description="Returns a new passkey registered with ubank.",
        epilog="You will be asked for your ubank password and secret code interactively. "
        "The passkey is encrypted with your ubank password.",
    )
    parser.add_argument("username", help="ubank username")
    parser.add_argument(
        "-o",
        "--output",
        default="-",
        type=argparse.FileType(mode="wb"),
        help="writes encrypted passkey to file (default: write to stdout)",
        dest="file",
    )
    parser.add_argument(
        "-n",
        "--passkey-name",
        default="ubank.py",
        help="sets passkey name (default: ubank.py)",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="displays httpx INFO logs"
    )
    args = parser.parse_args()
    if args.verbose:
        # Displays basic httpx request information.
        logging.basicConfig(level=logging.INFO)
    password = getpass("Enter ubank password: ")
    passkey = add_passkey(
        args.username,
        password=password,
        passkey_name=args.passkey_name,
    )
    passkey.dump(args.file, password=password)


if __name__ == "__main__":
    cli()



"""
ubank marimo notebook

Run:
    uv run marimo run notebook.py

Edit:
    uv run marimo edit notebook.py
"""

import marimo

__generated_with = "0.11.7"
app = marimo.App(width="medium")


@app.cell
def _():
    import marimo as mo

    return (mo,)


@app.cell
def _(mo):
    mo.md("""# ubank API Explo ~~it~~ rer""")
    return


@app.cell
def _(mo):
    uploaded_passkey = mo.ui.file(kind="area")
    username = mo.ui.text(placeholder="ubank username")
    password = mo.ui.text(placeholder="ubank password", kind="password")
    register = mo.ui.run_button(kind="success", label="Register")
    mo.md(
        f"""
        ## Authentication

        Access to ubank's API requires a passkey.
        """
    )
    return password, register, uploaded_passkey, username


@app.cell
def _(mo, password, register, uploaded_passkey, username):
    import io

    from cryptography.fernet import InvalidToken
    from httpx import HTTPStatusError

    from ubank import Passkey, add_passkey

    # stub out for testing
    # def add_passkey(username: str, password: str, passkey_name: str) -> Passkey:
    #     assert username
    #     assert passkey_name
    #     password = input("Enter security code ")
    #     return Passkey(passkey_name)

    new_passkey_file = io.BytesIO()
    if register.value:
        try:
            add_passkey(username.value, password.value, passkey_name="notebook.py").dump(
                new_passkey_file, password=password.value
            )
        except HTTPStatusError as e:
            print(f"Passkey registration failed: {e.__notes__[0]}")
    new_passkey = mo.download(
        new_passkey_file.getvalue(),
        filename="passkey.txt",
        label="passkey",
    )

    load_status = "⚪️"
    if uploaded_passkey.contents():
        try:
            Passkey.load(io.BytesIO(uploaded_passkey.contents()), password.value)
            load_status = "🟢"
        except InvalidToken:
            load_status = "🔴"

    mo.hstack(
        [
            mo.md(
                f"""
                ### Create passkey

                {register} a new {new_passkey} with ubank:

                {username} {password}


                You will be prompted for your security code.
                """
            ),
            mo.md(
                f"""
                ### Load passkey {load_status}

                Decrypt an existing passkey file with your {password}:

                {uploaded_passkey}
                """
            ),
        ],
        justify="start",
        gap=2.5,
    )
    return (
        HTTPStatusError,
        InvalidToken,
        Passkey,
        add_passkey,
        io,
        load_status,
        new_passkey,
        new_passkey_file,
    )


@app.cell
def _(mo):
    create_client = mo.ui.run_button(label="Create")

    mo.md(
        f"""
        ## API

        {create_client} an API client with your passkey.
        """
    )
    return (create_client,)


@app.cell
def _(
    InvalidToken,
    Passkey,
    create_client,
    io,
    mo,
    new_passkey_file,
    password,
    uploaded_passkey,
):
    from ubank import Api

    if create_client.value:
        try:
            with mo.status.spinner():
                passkey = Passkey.load(
                    # Coalesce bytes from uploaded or new passkey.
                    io.BytesIO(uploaded_passkey.contents() or new_passkey_file.getvalue()),
                    password.value,
                )
                api = Api(passkey)
                devices = api.get_devices(deviceUuid=passkey.device_id)
            mo.output.append(
                mo.md(
                    f"""
                    ### Devices

                    You've authenticated using passkey *{passkey.name}*.

                    The following security devices are registered with ubank:

                    {mo.ui.table([device.model_dump() for device in devices])}
                    """
                )
            )
        except InvalidToken:
            pass
    return Api, api, devices, passkey


@app.cell
def _(mo):
    get_balances = mo.ui.run_button(label="Get")

    mo.md(
        f"""
        ### Accounts

        {get_balances} account balances.
        """
    )
    return (get_balances,)


@app.cell
def _(api, get_balances, mo):
    if get_balances.value:
        with mo.status.spinner():
            accounts_response = api.get_accounts()
        mo.output.append(
            mo.tree(
                [
                    account.model_dump(include={"number", "label", "type", "balance"})
                    for account in accounts_response.linkedBanks[0].accounts
                ]
            )
        )
    return (accounts_response,)


@app.cell
def _(mo):
    from ubank import TransactionsSearchBody

    search_transactions = mo.ui.run_button(label="Search")
    from_date = mo.ui.date()
    to_date = mo.ui.date()
    limit = mo.ui.number(value=5)

    mo.md(
        f"""
        ### Transactions

        {search_transactions} for transactions between {from_date} and {to_date}.

        Limit the number of results to {limit}.
        """
    )
    return (
        TransactionsSearchBody,
        from_date,
        limit,
        search_transactions,
        to_date,
    )


@app.cell
def _(
    TransactionsSearchBody,
    api,
    from_date,
    mo,
    search_transactions,
    to_date,
):
    if search_transactions.value:
        with mo.status.spinner():
            search_response = api.post_accounts_transactions_search(
                body=TransactionsSearchBody(fromDate=from_date.value, toDate=to_date.value)
            )
        mo.output.append(
            mo.ui.table(
                search_response.model_dump(mode="json", exclude_none=True)["transactions"]
            )
        )
    return (search_response,)


if __name__ == "__main__":
    app.run()



from base64 import urlsafe_b64encode
from datetime import date

import pytest
from cryptography.hazmat.primitives import serialization
from httpx import HTTPStatusError

from ubank import (
    Api,
    Passkey,
    TransactionsSearchBody,
    UserVerifiedDevice,
    __version__,
    add_passkey,
    int8array_to_bytes,
    parse_public_key_credential_creation_options,
    parse_public_key_credential_request_options,
    prepare_assertion,
    prepare_attestation,
)


def test_version():
    assert __version__ == "2.1.0"


def test_int8array_to_bytes():
    """Tests bytes are correctly converted from Int8 to Uint8."""
    assert list(int8array_to_bytes([57, 18, 31, -66, -42, -43, -119, 90])) == [
        57,
        18,
        31,
        190,
        214,
        213,
        137,
        90,
    ]


def test_parse_public_key_credential_creation_options():
    """Tests options values are properly converted."""
    options = parse_public_key_credential_creation_options(
        '{"rp":{"id":"www.ubank.com.au","name":"www.ubank.com.au"},"user":{"id":[57,18,31,-66,-42,-43,-119,90],"displayName":"a663a471-01d67-4f5-aca3-2f9ae784c84d","name":"a663a471-01d67-4f5-aca3-2f9ae784c84d"},"challenge":[-128,-115,84,-82,90,35,20,-7,96,-121,-61,56,-59,93,-128,-37,-107,-41,37,62,-14,23,-102,23,-118,-76,-111,41,-88,36,59,36],"pubKeyCredParams":[{"type":"public-key","alg":"-7"},{"type":"public-key","alg":"-37"},{"type":"public-key","alg":"-257"}],"timeout":120000,"excludeCredentials":[{"type":"public-key","id":[57,18,31,-66,-42,-43,-119,91]}],"authenticatorSelection":{"authenticatorAttachment":"platform","residentKey":"required","requireResidentKey":true,"userVerification":"required"},"attestation":"none","extensions":{"credProps":true,"hmacCreateSecret":true}}'
    )
    assert options["publicKey"]["user"]["id"] == b"9\x12\x1f\xbe\xd6\xd5\x89Z"
    assert (
        options["publicKey"]["challenge"]
        == b"\x80\x8dT\xaeZ#\x14\xf9`\x87\xc38\xc5]\x80\xdb\x95\xd7%>\xf2\x17\x9a\x17\x8a\xb4\x91)\xa8$;$"
    )
    assert {"type": "public-key", "alg": -7} in options["publicKey"]["pubKeyCredParams"]
    # We captured the following challenge value from the wire sent back to ubank
    # (without the trailing '='). This verifies the conversion of the challenge
    # value from signed ints, to bytes and to base64 encoded bytes.
    assert (
        urlsafe_b64encode(options["publicKey"]["challenge"])
        == b"gI1UrlojFPlgh8M4xV2A25XXJT7yF5oXirSRKagkOyQ="
    )


def test_parse_public_key_credential_request_options():
    """Tests challenge value is properly converted."""
    options = parse_public_key_credential_request_options(
        '{"challenge":[-128,-115,84,-82,90,35,20,-7,96,-121,-61,56,-59,93,-128,-37,-107,-41,37,62,-14,23,-102,23,-118,-76,-111,41,-88,36,59,36],"timeout":120000,"rpId":"www.ubank.com.au","allowCredentials":[{"type":"public-key","id":[-50,90,62]},{"type":"public-key","id":[-29,89,87]}],"userVerification":"required"}'
    )
    assert (
        options["publicKey"]["challenge"]
        == b"\x80\x8dT\xaeZ#\x14\xf9`\x87\xc38\xc5]\x80\xdb\x95\xd7%>\xf2\x17\x9a\x17\x8a\xb4\x91)\xa8$;$"
    )


def test_serialize_attestation():
    """Tests bytes are converted to JSON-serializable strings."""
    attestation = {
        "id": b"VYVXb72FcRIDNpt2fYxxeCg62d_XBOdVoGqfSPGfJ5Q=",
        "rawId": b"U\x85Wo\xbd\x85q\x12\x036\x9bv}\x8cqx(:\xd9\xdf\xd7\x04\xe7U\xa0j\x9fH\xf1\x9f'\x94",
        "response": {
            "clientDataJSON": b'{"type": "webauthn.create", "challenge": "FPlgldclPvKHwzjFXYDbFw", "origin": "https://www.ubank.com.au"}',
            "attestationObject": b"\xa3cfmtdnonegattStmt\xa0hauthDataX\xa4\xf3\n\x01h\xc8E\xb5\xe5\x97>Q\x0c\xf2/\x0e\xfa\x80\x02\x9d\x15\xd6U\\\xab\x85\x9b\x0f\xbb\xfd%@^A\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 U\x85Wo\xbd\x85q\x12\x036\x9bv}\x8cqx(:\xd9\xdf\xd7\x04\xe7U\xa0j\x9fH\xf1\x9f'\x94\xa5\x01\x02\x03& \x01!X \xc2^\x9b\x11\x87\xc1\xb9\xa3t\xae\x0bR\xe4_\x189\xbf\xfa\xba\xc8=fy\xf6!h~7\xdb0\x8a1\"X \x93\xa7\xb3\x8d\xc4\x94\xa3\xadm\xb6\xc8_\x06\xbf]\xf6\xb6B\xae\x18^\xd3\x1e\xaaA\x00\x88\x05\x9a\xc1\x06B",
        },
        "type": "public-key",
    }

    assert prepare_attestation(attestation) == {
        "id": "VYVXb72FcRIDNpt2fYxxeCg62d_XBOdVoGqfSPGfJ5Q=",
        "rawId": "VYVXb72FcRIDNpt2fYxxeCg62d/XBOdVoGqfSPGfJ5Q=",
        "response": {
            "clientDataJSON": "eyJ0eXBlIjogIndlYmF1dGhuLmNyZWF0ZSIsICJjaGFsbGVuZ2UiOiAiRlBsZ2xkY2xQdktId3pqRlhZRGJGdyIsICJvcmlnaW4iOiAiaHR0cHM6Ly93d3cudWJhbmsuY29tLmF1In0=",
            "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVik8woBaMhFteWXPlEM8i8O+oACnRXWVVyrhZsPu/0lQF5BAAAAAAAAAAAAAAAAAAAAAAAAAAAAIFWFV2+9hXESAzabdn2McXgoOtnf1wTnVaBqn0jxnyeUpQECAyYgASFYIMJemxGHwbmjdK4LUuRfGDm/+rrIPWZ59iFofjfbMIoxIlggk6ezjcSUo61ttshfBr9d9rZCrhhe0x6qQQCIBZrBBkI=",
        },
        "type": "public-key",
    }


def test_serialize_assertion():
    """Tests bytes are converted to JSON-serializable strings."""
    assertion = {
        "id": b"AI71FKYn9UU0GlDFWF03d7LKJu_cHdo3gJUJRao39bQ=",
        "rawId": b"\x00\x8e\xf5\x14\xa6'\xf5E4\x1aP\xc5X]7w\xb2\xca&\xef\xdc\x1d\xda7\x80\x95\tE\xaa7\xf5\xb4",
        "response": {
            "authenticatorData": b"\xf3\n\x01h\xc8E\xb5\xe5\x97>Q\x0c\xf2/\x0e\xfa\x80\x02\x9d\x15\xd6U\\\xab\x85\x9b\x0f\xbb\xfd%@^\x01\x00\x00\x00\x01",
            "clientDataJSON": b'{"type": "webauthn.get", "challenge": "gI1UrlojFPlgh8M4xV2A25XXJT7yF5oXirSRKagkOyQ", "origin": "https://www.ubank.com.au"}',
            "signature": b"0E\x02!\x00\xe0\x12>f\xe3\x11\xac\x02n\x1e\x05\xc80\x97\x99L\x89\xf6H0\x0e7qo\xb0\xafft\xf6C=\xac\x02 \x0cT\x0b @Q\xf3*\xe9-.#\x88EE\r\x16\xa7S\xbd\xd0\xbd\xb7f\xcc\xd2t\x0f \xa2/\xc8",
            "userHandle": b"9\x12\x1f\xbe\xd6\xd5\x89Z",
        },
        "type": "public-key",
    }

    assert prepare_assertion(assertion) == {
        "id": "AI71FKYn9UU0GlDFWF03d7LKJu_cHdo3gJUJRao39bQ=",
        "rawId": "AI71FKYn9UU0GlDFWF03d7LKJu/cHdo3gJUJRao39bQ=",
        "response": {
            "authenticatorData": "8woBaMhFteWXPlEM8i8O+oACnRXWVVyrhZsPu/0lQF4BAAAAAQ==",
            "clientDataJSON": "eyJ0eXBlIjogIndlYmF1dGhuLmdldCIsICJjaGFsbGVuZ2UiOiAiZ0kxVXJsb2pGUGxnaDhNNHhWMkEyNVhYSlQ3eUY1b1hpclNSS2Fna095USIsICJvcmlnaW4iOiAiaHR0cHM6Ly93d3cudWJhbmsuY29tLmF1In0=",
            "signature": "MEUCIQDgEj5m4xGsAm4eBcgwl5lMifZIMA43cW+wr2Z09kM9rAIgDFQLIEBR8yrpLS4jiEVFDRanU73QvbdmzNJ0DyCiL8g=",
            "userHandle": "ORIfvtbViVo=",
        },
        "type": "public-key",
    }


def test_passkey_serialization(tmp_path):
    """Tests passkey serialization."""
    passkey = Passkey(name="")
    passkey.soft_webauthn_device.create(
        {
            "publicKey": {
                "rp": {"name": "example org", "id": "example.org"},
                "user": {
                    "id": b"randomhandle",
                    "name": "username",
                    "displayName": "user name",
                },
                "challenge": b"arandomchallenge",
                "pubKeyCredParams": [{"alg": -7, "type": "public-key"}],
                "attestation": "none",
            }
        },
        "https://example.org",
    )

    # Serialize and deserialize passkey.
    with (tmp_path / "passkey.txt").open("wb") as f:
        passkey.dump(f, password="")
    with (tmp_path / "passkey.txt").open("rb") as f:
        deserialized_passkey = Passkey.load(f, password="")
    assert type(deserialized_passkey.soft_webauthn_device) is UserVerifiedDevice


def test_add_passkey_bad_username():
    """HTTPStatusErrors raised by add_passkey() should contain notes with API error text."""
    try:
        add_passkey("badusername", "password", "")
    except HTTPStatusError as e:
        assert e.__notes__
        assert "Username invalid" in e.__notes__[0]


def test_api():
    """Tests declared API methods using passkey loaded from passkey.txt."""
    # Skip test if hard-coded passkey file is not available.
    try:
        with open("passkey.txt", "rb") as f:
            passkey = Passkey.load(f)
    except FileNotFoundError as e:
        pytest.skip(str(e))

    # Authenticate to ubank with passkey.
    with Api(passkey) as api:
        assert (
            api.client.get("accounts").json()["linkedBanks"][0]["accounts"][0][
                "balance"
            ]["currency"]
            == "AUD"
        )

        customer = api.get_customer_details()
        assert customer.addresses[0].addressFormat == "AUS"
        accounts = api.get_accounts()
        assert accounts.linkedBanks[0].accounts[0].balance.currency == "AUD"
        assert (
            api.get_accounts_summary().linkedBanks[0].accounts[0].balance.currency
            == "AUD"
        )
        assert api.post_accounts_transactions_search(
            body=TransactionsSearchBody(
                fromDate=date(2024, 1, 1), toDate=date(2025, 1, 1), limit=100
            )
        )
        for account in accounts.linkedBanks[0].accounts:
            assert api.get_account_transactions(
                account_id=account.id,
                bank_id=accounts.linkedBanks[0].bankId,
                customerId=customer.customerId,
            )
        assert api.get_cards()
        # Name of a device should match this passkey.
        assert [
            device
            for device in api.get_devices(deviceUuid=passkey.device_id)
            if device.deviceName == passkey.name
        ]
        assert api.delete_device(device_id="test-2f9ae784c84d")
        assert api.get_contacts()
